!> @file vdi_internal_controls.f90
!--------------------------------------------------------------------------------!
! This file is part of the PALM model system.
!
! PALM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! PALM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! PALM. If not, see <http://www.gnu.org/licenses/>.
!
! Copyright 1997-2019 Leibniz Universitaet Hannover
!--------------------------------------------------------------------------------!
!
! Current revisions:
! -----------------
!
!
! Former revisions:
! -----------------
! 3986 2019-06-03 07:25:57Z weniger
!
! Description:
! ------------
!> According to VDI Guideline 3783 Part 9, internal assessment have to be
!> carried out within the program for the model to be considered as evaluated.
!------------------------------------------------------------------------------!
 MODULE vdi_internal_controls

    USE arrays_3d


    USE control_parameters

    USE indices


    USE kinds

    USE pegrid

    IMPLICIT NONE

      INTEGER(iwp) ::  x_cgp = 0
      INTEGER(iwp) ::  y_cgp = 0
      INTEGER(iwp) ::  z_cgp = 0

    INTERFACE vdi_2_deltat_wave
       MODULE PROCEDURE vdi_2_deltat_wave
    END INTERFACE vdi_2_deltat_wave

    INTERFACE vdi_standard_differences
       MODULE PROCEDURE vdi_standard_differences
    END INTERFACE vdi_standard_differences

    INTERFACE vdi_domain_averages
       MODULE PROCEDURE vdi_domain_averages
    END INTERFACE vdi_domain_averages

    INTERFACE vdi_plausible_values
       MODULE PROCEDURE vdi_plausible_values
    END INTERFACE vdi_plausible_values

    INTERFACE vdi_actions
       MODULE PROCEDURE vdi_actions
    END INTERFACE vdi_actions

    INTERFACE vdi_conservation_of_mass
       MODULE PROCEDURE vdi_conservation_of_mass
    END INTERFACE vdi_conservation_of_mass


    SAVE

    PRIVATE

!
!-- Public functions / NEEDS SORTING
    PUBLIC vdi_conservation_of_mass,  &
           vdi_plausible_values, &
           vdi_actions,      &
           vdi_2_deltat_wave, &
           vdi_standard_differences, &
           vdi_domain_averages

 CONTAINS

!------------------------------------------------------------------------------!
! Description:
! ------------
!> Call for all grid points
!------------------------------------------------------------------------------!
    SUBROUTINE vdi_actions( location )


    CHARACTER (LEN=*), INTENT(IN) ::  location !< call location string

    SELECT CASE ( location )

       CASE ( 'after_integration' )

          CALL vdi_2_deltat_wave

          CALL vdi_standard_differences

          CALL vdi_domain_averages

           CALL vdi_conservation_of_mass

           CALL vdi_plausible_values


       CASE DEFAULT
          CONTINUE

    END SELECT

    END SUBROUTINE vdi_actions
!------------------------------------------------------------------------------!
! Description:
! ------------
!> At a control grid point in the interior of the model domain,
!> 2 * deltat waves are not to be generated with increasing simulation time.
!------------------------------------------------------------------------------!
   SUBROUTINE vdi_2_deltat_wave

      USE arrays_3d

      USE control_parameters

      USE indices

      USE kinds


      IMPLICIT NONE


      INTEGER(iwp) :: count_wave = 0
      INTEGER(iwp) :: count_time = 0

      REAL(wp) ::  rr


      INTEGER(iwp), DIMENSION(4) ::  sig_arr = (/ 0, 0, 0, 0/)


!
!--   Defining the control grid point
      IF ( time_since_reference_point <= dt_3d )  THEN
         x_cgp = INT(nxl + (nxr-nxl)/2)
         y_cgp = INT(nys + (nyn-nys)/2)
         z_cgp = INT(nzt/2)
!
!--      If the grid point lies in a building, a new point is defined
         DO WHILE (.NOT. BTEST( wall_flags_0(z_cgp,y_cgp,x_cgp), 1) )
            CALL RANDOM_NUMBER(rr)
            z_cgp = z_cgp+FLOOR((nzt-1)*rr)   !Random number upon z_cgp
         ENDDO

      ENDIF



      IF ( u_p(z_cgp,y_cgp,x_cgp) - u(z_cgp,y_cgp,x_cgp) > 0 )  THEN
         sig_arr(4) = 0
      ELSE
         sig_arr(4) = 1
      ENDIF


!
!--   With this criterion 2 delta t waves are detected if the amplitude of
!--   the waves is greater than the change of u with increasing time
      IF ( sig_arr(1) == 1 .AND. sig_arr(2) == 0 .AND. sig_arr(3) == 0 .AND.   &
                                                 sig_arr(4) == 1 )  THEN

         count_wave = count_wave + 1

         IF ( count_wave == 4 .AND. count_time == 4 )  THEN
            message_string = '2 deltat waves are generated '
            CALL message( 'vdi_2_deltat_wave', 'PAXXXX', 0, 2, myid, 6, 0 )  
         ENDIF

         count_time = 0
      ELSE
         IF ( count_time >= 4 )  THEN
            count_wave = 0
         ENDIF
      ENDIF


      sig_arr(1) = sig_arr(2)
      sig_arr(2) = sig_arr(3)
      sig_arr(3) = sig_arr(4)


      count_time = count_time + 1

   END SUBROUTINE vdi_2_deltat_wave

!------------------------------------------------------------------------------!
! Description:
! ------------
!> In this internal assessment the maxima of standarddifferences of the
!> meteorological variables, computed layer by layer will be checked.
!> The maxima should not to remain at the open edges of the model or
!> travel from there into the interior of the domain with increasing
!> simulation time.
!------------------------------------------------------------------------------!
   SUBROUTINE vdi_standard_differences

      USE arrays_3d,                                                           &
          ONLY:  u, v, w, q, pt

      USE control_parameters

      USE indices

      USE kinds

      USE pegrid


      IMPLICIT NONE


      INTEGER(iwp) ::  pos(1)
      INTEGER(iwp) ::  position_u_diff = 0
      INTEGER(iwp) ::  position_u_diff_p = 0
      INTEGER(iwp) ::  position_u_diff_pp = 0
      INTEGER(iwp) ::  position_v_diff = 0
      INTEGER(iwp) ::  position_v_diff_p = 0
      INTEGER(iwp) ::  position_v_diff_pp = 0
      INTEGER(iwp) ::  position_w_diff = 0
      INTEGER(iwp) ::  position_w_diff_p = 0
      INTEGER(iwp) ::  position_w_diff_pp = 0
      INTEGER(iwp) ::  position_q_diff = 0
      INTEGER(iwp) ::  position_q_diff_p = 0
      INTEGER(iwp) ::  position_q_diff_pp = 0
      INTEGER(iwp) ::  position_pt_diff = 0
      INTEGER(iwp) ::  position_pt_diff_p = 0
      INTEGER(iwp) ::  position_pt_diff_pp = 0


      REAL(wp), DIMENSION(nzb:nzt+1) ::  u_diff
      REAL(wp), DIMENSION(nzb:nzt+1) ::  v_diff
      REAL(wp), DIMENSION(nzb:nzt+1) ::  w_diff
      REAL(wp), DIMENSION(nzb:nzt+1) ::  q_diff
      REAL(wp), DIMENSION(nzb:nzt+1) ::  pt_diff



!
!--   Calculation of the standard differences of u
      CALL stand_diff(u,u_diff,1)

!
!--   Determination of the position of the maximum
      pos = MAXLOC(u_diff)
      position_u_diff = pos(1)

!
!--   Check the position of the maximum of the standard differences of u
      IF ( time_since_reference_point > 2* dt_3d )  THEN
         CALL check_position(position_u_diff,position_u_diff_p,                &
                                                position_u_diff_pp)
      ENDIF

      position_u_diff_pp = position_u_diff_p
      position_u_diff_p = position_u_diff

!
!--   Calculation of the standard differences of v
      CALL stand_diff(v,v_diff,2)

!
!--   Determination of the position of the maximum
      pos = MAXLOC(v_diff)
      position_v_diff = pos(1)


!
!--   Check the position of the maximum of the standard differences of v
      IF ( time_since_reference_point > 2* dt_3d )  THEN
         CALL check_position(position_v_diff,position_v_diff_p,                &
                                              position_v_diff_pp)
      ENDIF

      position_v_diff_pp = position_v_diff_p
      position_v_diff_p = position_v_diff


!
!--   Calculation of the standard differences of w
      CALL stand_diff(w,w_diff,3)

!
!--   Determination of the position of the maximum
      pos = MAXLOC(w_diff)
      position_w_diff = pos(1)

!
!--   Check the position of the maximum of the standard differences of w
      IF ( time_since_reference_point > 2* dt_3d )  THEN
         CALL check_position(position_w_diff,position_w_diff_p,                &
                                              position_w_diff_pp)
      ENDIF

      position_w_diff_pp = position_w_diff_p
      position_w_diff_p = position_w_diff


!
!--   Calculation of the standard differences of q
      IF ( humidity )  THEN
         CALL stand_diff(q,q_diff,0)

!
!--      Determination of the position of the maximum
         pos = MAXLOC(q_diff)
         position_q_diff = pos(1)
!                      position_q_diff = 5

!
!--      Check the position of the maximum of the standard differences of q
         IF ( time_since_reference_point > 2* dt_3d )  THEN
            CALL check_position(position_q_diff,position_q_diff_p,       &
                                                    position_q_diff_pp)
         ENDIF

         position_q_diff_pp = position_q_diff_p
         position_q_diff_p = position_q_diff

      ENDIF


!
!--   Calculation of the standard differences of pt
      IF ( .NOT. neutral )  THEN
         CALL stand_diff(pt,pt_diff,0)
!
!--      Determination of the position of the maximum
         pos = MAXLOC(pt_diff)
         position_pt_diff = pos(1)

!
!--      Check the position of the maximum of the standard differences of pt
         IF ( time_since_reference_point > 2* dt_3d )  THEN
            CALL check_position(position_pt_diff,position_pt_diff_p,           &
                                                   position_pt_diff_pp)
         ENDIF

         position_pt_diff_pp = position_pt_diff_p
         position_pt_diff_p = position_pt_diff

      ENDIF

   END SUBROUTINE vdi_standard_differences


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Calculation of the standard deviation
!------------------------------------------------------------------------------!


   SUBROUTINE stand_diff(quant,std_diff,quant_type)

      USE pegrid

      IMPLICIT NONE

      INTEGER(iwp) ::  i
      INTEGER(iwp) ::  j
      INTEGER(iwp) ::  k
      INTEGER(iwp) ::  quant_type

      INTEGER(iwp), DIMENSION(nzb:nzt+1) ::  count_2d_l
      INTEGER(iwp), DIMENSION(nzb:nzt+1) ::  count_2d

      REAL(wp), DIMENSION(nzb:nzt+1) ::  quant_av_k_l
      REAL(wp), DIMENSION(nzb:nzt+1) ::  quant_av_k
      REAL(wp), DIMENSION(nzb:nzt+1) ::  std_diff
      REAL(wp), DIMENSION(nzb:nzt+1) ::  std_diff_l
      REAL(wp), DIMENSION(nzb:nzt+1,nysg:nyng,nxlg:nxrg) ::  quant


!
!--   Calculation of the standard differences
      quant_av_k_l = 0.0_wp
      quant_av_k = 0.0_wp
      std_diff = 0.0_wp
      std_diff_l = 0.0_wp



      !average
      count_2d_l = 0
      count_2d = 0
      DO  i = nxl, nxr
         DO  j = nys, nyn
            DO  k = nzb+1, nzt+1
               IF ( BTEST( wall_flags_0(k,j,i), quant_type) ) THEN
                  quant_av_k_l(k) = quant_av_k_l(k) + quant(k,j,i)
                  count_2d_l(k) = count_2d_l(k) + 1
               ENDIF
            ENDDO
         ENDDO
      ENDDO



#if defined( __parallel )
      CALL MPI_ALLREDUCE( quant_av_k_l, quant_av_k, nzt+1-nzb+1,               &
                               MPI_REAL, MPI_SUM, comm2d, ierr )
#else
      quant_av_k = quant_av_k_l
#endif



#if defined( __parallel )
      CALL MPI_ALLREDUCE( count_2d_l, count_2d, nzt+1-nzb+1,                   &
                                 MPI_INTEGER, MPI_SUM, comm2d, ierr )
#else
      count_2d = count_2d_l
#endif


      DO  k = nzb+1, nzt+1
         quant_av_k(k) = quant_av_k(k) / REAL( count_2d(k), KIND=wp )
      ENDDO


      DO  i = nxl, nxr
         DO  j = nys, nyn
            DO  k = nzb+1, nzt+1
               IF ( BTEST( wall_flags_0(k,j,i), quant_type) ) THEN
                  std_diff_l(k) = std_diff_l(k) + (quant(k,j,i) -              &
                                                   quant_av_k(k))**2
               ENDIF
            ENDDO
         ENDDO
      ENDDO


#if defined( __parallel )
      CALL MPI_ALLREDUCE( std_diff_l, std_diff, nzt+1-nzb+1,                   &
                               MPI_REAL, MPI_SUM, comm2d, ierr )
#else
      std_diff = std_diff_l
#endif

      DO  k = nzb+1, nzt+1
         std_diff(k) = SQRT(std_diff(k) / REAL( count_2d(k), KIND=wp ))
      ENDDO




   END SUBROUTINE stand_diff

!------------------------------------------------------------------------------!
! Description:
! ------------
!> Tests for the position of the maxima of the standard deviation
!------------------------------------------------------------------------------!


   SUBROUTINE check_position(position_std_diff,position_std_diff_p,            &
                                                   position_std_diff_pp)

      IMPLICIT NONE

      INTEGER(iwp) ::  position_std_diff
      INTEGER(iwp) ::  position_std_diff_p
      INTEGER(iwp) ::  position_std_diff_pp


      IF ( position_std_diff == nzt .AND. position_std_diff_p == nzt .AND. &
                                         position_std_diff_pp == nzt)  THEN
         message_string = 'The maxima of the standard differences remain '//   &
                          'at the open edges of the model.'
         CALL message( 'vdi_standard_differences', 'PAXXXX', 0, 2, 0, 6, 0 )
      ENDIF

      IF ( position_std_diff == nzt-2 .AND. position_std_diff_p == nzt-1 .AND. &
                                        position_std_diff_pp == nzt )  THEN
         message_string = 'The maxima of the standard differences travel '//   &
                          'from the open edges into the interior '//           &
                          'of the domain with increasing simulation time.'
         CALL message( 'vdi_standard_differences', 'PAXXXX', 0, 2, 0, 6, 0 )
      ENDIF


   END SUBROUTINE check_position


!------------------------------------------------------------------------------!
! Description:
! ------------
!> In this control it will be checked, if the means of the meteorological
!> variables over the model grid are not to exhibit 2 deltat waves or
!> monotonic increase or decrease with increasing simulation time.
!------------------------------------------------------------------------------!
   SUBROUTINE vdi_domain_averages

      USE arrays_3d

      USE control_parameters

      USE indices

      USE kinds

      USE pegrid


      IMPLICIT NONE

      INTEGER(iwp) ::  i
      INTEGER(iwp) ::  j
      INTEGER(iwp) ::  k
      INTEGER(iwp) :: mono_count_u = 0
      INTEGER(iwp) :: mono_count_v = 0
      INTEGER(iwp) :: mono_count_w = 0
      INTEGER(iwp) :: mono_count_q = 0
      INTEGER(iwp) :: mono_count_pt = 0
      INTEGER(iwp) :: count_wave_u = 0
      INTEGER(iwp) :: count_wave_v = 0
      INTEGER(iwp) :: count_wave_w = 0
      INTEGER(iwp) :: count_wave_q = 0
      INTEGER(iwp) :: count_wave_pt = 0
      INTEGER(iwp) :: count_time_u = 0
      INTEGER(iwp) :: count_time_v = 0
      INTEGER(iwp) :: count_time_w = 0
      INTEGER(iwp) :: count_time_q = 0
      INTEGER(iwp) :: count_time_pt = 0

      REAL(wp) ::  u_av_l = 0.0_wp
      REAL(wp) ::  u_av = 0.0_wp
      REAL(wp) ::  u_av_p = 0.0_wp
      REAL(wp) ::  v_av_l = 0.0_wp
      REAL(wp) ::  v_av = 0.0_wp
      REAL(wp) ::  v_av_p = 0.0_wp
      REAL(wp) ::  w_av_l = 0.0_wp
      REAL(wp) ::  w_av = 0.0_wp
      REAL(wp) ::  w_av_p = 0.0_wp
      REAL(wp) ::  q_av_l = 0.0_wp
      REAL(wp) ::  q_av = 0.0_wp
      REAL(wp) ::  q_av_p = 0.0_wp
      REAL(wp) ::  pt_av_l = 0.0_wp
      REAL(wp) ::  pt_av = 0.0_wp
      REAL(wp) ::  pt_av_p = 0.0_wp
      REAL(wp) ::  average_count = 0.0_wp
      REAL(wp) ::  average_count_l = 0.0_wp
      REAL(wp) ::  w_rel = 0.0_wp



      INTEGER(iwp), DIMENSION(4) ::  sig_u_arr = (/ 0, 0, 0, 0/)
      INTEGER(iwp), DIMENSION(4) ::  sig_v_arr = (/ 0, 0, 0, 0/)
      INTEGER(iwp), DIMENSION(4) ::  sig_w_arr = (/ 0, 0, 0, 0/)
      INTEGER(iwp), DIMENSION(4) ::  sig_q_arr = (/ 0, 0, 0, 0/)
      INTEGER(iwp), DIMENSION(4) ::  sig_pt_arr = (/ 0, 0, 0, 0/)


!
!--   Averaging the meteorological variables over the model grid

      average_count_l = 0.0_wp
      u_av_l = 0.0_wp
      DO  i = nxl, nxr
         DO  j = nys, nyn
            DO  k = nzb, nzt+1
               IF ( BTEST( wall_flags_0(k,j,i), 1) ) THEN
                  u_av_l = u_av_l + u(k,j,i)
                  average_count_l = average_count_l + 1.
               ENDIF
            ENDDO
         ENDDO
      ENDDO


#if defined( __parallel )
      CALL MPI_ALLREDUCE( u_av_l, u_av, 1,                                     &
                             MPI_REAL, MPI_SUM, comm2d, ierr )
#else
      u_av = u_av_l
#endif



#if defined( __parallel )
      CALL MPI_ALLREDUCE( average_count_l, average_count, 1,                   &
                                 MPI_REAL, MPI_SUM, comm2d, ierr )
#else
      average_count = average_count_l
#endif



      u_av = u_av / average_count



      average_count = 0.0_wp
      v_av_l = 0.0_wp
      DO  i = nxl, nxr
         DO  j = nys, nyn
            DO  k = nzb, nzt+1
               IF ( BTEST( wall_flags_0(k,j,i), 2) ) THEN
                  v_av_l = v_av_l + v(k,j,i)
                  average_count_l = average_count_l + 1.
               ENDIF
            ENDDO
         ENDDO
      ENDDO

#if defined( __parallel )
      CALL MPI_ALLREDUCE( v_av_l, v_av, 1,                                     &
                             MPI_REAL, MPI_SUM, comm2d, ierr )
#else
      v_av = v_av_l
#endif



#if defined( __parallel )
      CALL MPI_ALLREDUCE( average_count_l, average_count, 1,                   &
                                 MPI_REAL, MPI_SUM, comm2d, ierr )
#else
      average_count = average_count_l
#endif

      v_av = v_av/average_count


      average_count = 0.0_wp
      w_av_l = 0.0_wp
      DO  i = nxl, nxr
         DO  j = nys, nyn
            DO  k = nzb, nzt+1
               IF ( BTEST( wall_flags_0(k,j,i), 3) ) THEN
                  w_av_l = w_av_l + w(k,j,i)
                  average_count_l = average_count_l + 1.
               ENDIF
            ENDDO
         ENDDO
      ENDDO

#if defined( __parallel )
      CALL MPI_ALLREDUCE( w_av_l, w_av, 1,                                     &
                             MPI_REAL, MPI_SUM, comm2d, ierr )
#else
      w_av = w_av_l
#endif



#if defined( __parallel )
      CALL MPI_ALLREDUCE( average_count_l, average_count, 1,                   &
                                 MPI_REAL, MPI_SUM, comm2d, ierr )
#else
      average_count = average_count_l
#endif

      w_av = w_av/average_count


      IF ( humidity )  THEN
         average_count = 0.0_wp
         q_av_l = 0.0_wp
         DO  i = nxl, nxr
            DO  j = nys, nyn
               DO  k = nzb, nzt+1
                  IF ( BTEST( wall_flags_0(k,j,i), 0) ) THEN
                     q_av_l = q_av_l + q(k,j,i)
                     average_count_l = average_count_l + 1.
                  ENDIF
               ENDDO
            ENDDO
         ENDDO


#if defined( __parallel )
         CALL MPI_ALLREDUCE( q_av_l, q_av, 1,                                  &
                             MPI_REAL, MPI_SUM, comm2d, ierr )
#else
         q_av = q_av_l
#endif



#if defined( __parallel )
         CALL MPI_ALLREDUCE( average_count_l, average_count, 1,                &
                                 MPI_REAL, MPI_SUM, comm2d, ierr )
#else
         average_count = average_count_l
#endif

         q_av = q_av/average_count
      ENDIF


      IF ( .NOT. neutral )  THEN
         average_count = 0.0_wp
         pt_av_l = 0.0_wp
         DO  i = nxl, nxr
            DO  j = nys, nyn
               DO  k = nzb, nzt+1
                  IF ( BTEST( wall_flags_0(k,j,i), 0) ) THEN
                     pt_av_l = pt_av_l + pt(k,j,i)
                     average_count_l = average_count_l + 1.
                  ENDIF
               ENDDO
            ENDDO
         ENDDO

#if defined( __parallel )
         CALL MPI_ALLREDUCE( pt_av_l, pt_av, 1,                                &
                             MPI_REAL, MPI_SUM, comm2d, ierr )
#else
         pt_av = pt_av_l
#endif



#if defined( __parallel )
         CALL MPI_ALLREDUCE( average_count_l, average_count, 1,                &
                                 MPI_REAL, MPI_SUM, comm2d, ierr )
#else
         average_count = average_count_l
#endif

         pt_av = pt_av/average_count
      ENDIF

!
!--   Determination of the difference to the previous time step and the sign
      IF ( time_since_reference_point > dt_3d )  THEN


         IF ( u_av_p - u_av > 0 )  THEN
            sig_u_arr(4) = 0
         ELSE
            sig_u_arr(4) = 1
         ENDIF

         IF ( v_av_p - v_av > 0 )  THEN
            sig_v_arr(4) = 0
         ELSE
            sig_v_arr(4) = 1
         ENDIF


         IF ( w_av_p - w_av > 0 )  THEN
            sig_w_arr(4) = 0
         ELSE
            sig_w_arr(4) = 1
         ENDIF

         IF ( humidity )  THEN
            IF ( q_av_p - q_av > 0 )  THEN
               sig_q_arr(4) = 0
            ELSE
               sig_q_arr(4) = 1
            ENDIF
         ENDIF

         IF ( .NOT. neutral )  THEN
            IF ( pt_av_p - pt_av > 0 )  THEN
               sig_pt_arr(4) = 0
            ELSE
               sig_pt_arr(4) = 1
            ENDIF
         ENDIF


      ENDIF



!
!--   Testing for 2 deltat waves
      IF ( sig_u_arr(1) == 1 .AND. sig_u_arr(2) == 0 .AND. sig_u_arr(3) == 0   &
                                           .AND. sig_u_arr(4) == 1 )  THEN
         count_wave_u = count_wave_u + 1

         IF ( count_wave_u == 4 .AND. count_time_u == 4 )  THEN  !< if there are 4 waves in a row, the simulation should be aborted
            message_string = '2 deltat waves are generated '
            CALL message( 'vdi_domain_averages', 'PAXXXX', 0, 2, 0, 6, 0 )
         ENDIF
      count_time_u = 0
      ELSE
         IF ( count_time_u >= 4 )  THEN
            count_wave_u = 0
         ENDIF
      ENDIF


      IF ( sig_v_arr(1) == 1 .AND. sig_v_arr(2) == 0 .AND. sig_v_arr(3) == 0   &
                                           .AND. sig_v_arr(4) == 1 )  THEN
         count_wave_v = count_wave_v + 1

         IF ( count_wave_v == 4 .AND. count_time_v == 4 )  THEN
            message_string = '2 deltat waves are generated '
            CALL message( 'vdi_domain_averages', 'PAXXXX', 0, 2, 0, 6, 0 )
         ENDIF

         count_time_v = 0
      ELSE
         IF ( count_time_v >= 4 )  THEN
            count_wave_v = 0
         ENDIF
      ENDIF


        w_rel = ABS(w_av_p - w_av)/w_av  !< Calculation of the rel. deviation
!--
!--   in the case of the w component, the relative amplitude of the waves must not be greater than 1 %
      IF ( sig_w_arr(1) == 1 .AND. sig_w_arr(2) == 0 .AND. sig_w_arr(3) == 0   &
                    .AND. sig_w_arr(4) == 1 .AND. w_rel > 0.01 )  THEN

         count_wave_w = count_wave_w + 1

         IF ( count_wave_w == 4 .AND. count_time_w == 4 )  THEN
            message_string = '2 deltat waves are generated '
            CALL message( 'vdi_domain_averages', 'PAXXXX', 0, 2, 0, 6, 0 )
         ENDIF

         count_time_w = 0
      ELSE
        IF ( count_time_w >= 4 )  THEN
           count_wave_w = 0
        ENDIF
      ENDIF

      IF ( humidity )  THEN
         IF ( sig_q_arr(1) == 1 .AND. sig_q_arr(2) == 0 .AND. sig_q_arr(3) == 0&
                                           .AND. sig_q_arr(4) == 1 )  THEN

            count_wave_q = count_wave_q + 1

            IF ( count_wave_q == 4 .AND. count_time_q  == 4 )  THEN
               message_string = '2 deltat waves are generated '
               CALL message( 'vdi_domain_averages', 'PAXXXX', 0, 2, 0, 6, 0 )
            ENDIF

            count_time_q = 0
         ELSE
            IF ( count_time_q >= 4 )  THEN
               count_wave_q = 0
            ENDIF
         ENDIF
      ENDIF



      IF ( .NOT. neutral ) THEN
         IF ( sig_pt_arr(1) == 1 .AND. sig_pt_arr(2) == 0                      &
                .AND. sig_pt_arr(3) == 0  .AND. sig_pt_arr(4) == 1 )  THEN

            count_wave_pt = count_wave_pt + 1

            IF ( count_wave_pt == 4 .AND. count_time_pt == 4 )  THEN
               message_string = '2 deltat waves are generated '
               CALL message( 'vdi_domain_averages', 'PAXXXX', 0, 2, 0, 6, 0 )
            ENDIF

            count_time_pt = 0
         ELSE
           IF ( count_time_pt >= 4 )  THEN
              count_wave_pt = 0
           ENDIF
         ENDIF
      ENDIF





!
!--   Testing if there is a monotonic increase or decrease with increasing simulation time
      IF ( (sig_u_arr(3) == 0 .AND. sig_u_arr(4) == 1) .OR. (sig_u_arr(3) == 1 &
                                            .AND. sig_u_arr(4) == 0) )  THEN
         mono_count_u = 0
      ELSE
         mono_count_u = mono_count_u + 1
      ENDIF



      IF ( time_since_reference_point >= end_time .AND. mono_count_u > 0.9_wp  &
                                         * current_timestep_number ) THEN
         message_string = 'Monotonic decrease or increase with ' //            &
                          'increasing simulation time for u'
         CALL message( 'vdi_domain_averages', 'PAXXXX', 0, 1, 0, 6, 0 )
      ENDIF


      IF ( (sig_v_arr(3) == 0 .AND. sig_v_arr(4) == 1) .OR. (sig_v_arr(3) == 1 &
                                           .AND. sig_v_arr(4) == 0) )  THEN
         mono_count_v = 0
      ELSE
         mono_count_v = mono_count_v + 1
      ENDIF


      IF ( time_since_reference_point >= end_time .AND. mono_count_v > 0.9_wp  &
                                          * current_timestep_number ) THEN
         message_string = 'Monotonic decrease or increase with ' //            &
                          'increasing simulation time for v'
         CALL message( 'vdi_domain_averages', 'PAXXXX', 0, 1, 0, 6, 0 )
      ENDIF



      IF ( (sig_w_arr(3) == 0 .AND. sig_w_arr(4) == 1) .OR. (sig_w_arr(3) == 1 &
                                         .AND. sig_w_arr(4) == 0) )  THEN
         mono_count_w = 0
      ELSE
         mono_count_w = mono_count_w + 1
      ENDIF


      IF ( time_since_reference_point >= end_time .AND. mono_count_w > 0.9_wp  &
                                         * current_timestep_number ) THEN
         message_string = 'Monotonic decrease or increase with ' //            &
                          'increasing simulation time for w'
         CALL message( 'vdi_domain_averages', 'PAXXXX', 0, 1, 0, 6, 0 )
      ENDIF


      IF ( humidity )  THEN
         IF ( (sig_q_arr(3) == 0 .AND. sig_q_arr(4) == 1) .OR.                 &
                 (sig_q_arr(3) == 1 .AND. sig_q_arr(4) == 0) )  THEN
            mono_count_q = 0
         ELSE
            mono_count_q = mono_count_q + 1
         ENDIF


         IF ( time_since_reference_point >= end_time .AND.                     &
                     mono_count_q > 0.9_wp * current_timestep_number ) THEN
            message_string = 'Monotonic decrease or increase with ' //         &
                             'increasing simulation time for q'
            CALL message( 'vdi_domain_averages', 'PAXXXX', 0, 1, 0, 6, 0 )
         ENDIF
      ENDIF


      IF ( .NOT. neutral )  THEN
         IF ( (sig_pt_arr(3) == 0 .AND. sig_pt_arr(4) == 1) .OR.               &
                     (sig_pt_arr(3) == 1 .AND. sig_pt_arr(4) == 0) )  THEN
            mono_count_pt = 0
         ELSE
            mono_count_pt = mono_count_pt + 1
         ENDIF


         IF ( time_since_reference_point >= end_time .AND. mono_count_pt > 0.9_wp          &
                                         * current_timestep_number ) THEN
            message_string = 'Monotonic decrease or increase with ' //         &
                             'increasing simulation time for pt'
            CALL message( 'vdi_domain_averages', 'PAXXXX', 0, 1, 0, 6, 0 )
         ENDIF
      ENDIF


!
!--   Save the values from the previous time step
      sig_u_arr(1) = sig_u_arr(2)
      sig_u_arr(2) = sig_u_arr(3)
      sig_u_arr(3) = sig_u_arr(4)

      u_av_p = u_av


      sig_v_arr(1) = sig_v_arr(2)
      sig_v_arr(2) = sig_v_arr(3)
      sig_v_arr(3) = sig_v_arr(4)

      v_av_p = v_av


      sig_w_arr(1) = sig_w_arr(2)
      sig_w_arr(2) = sig_w_arr(3)
      sig_w_arr(3) = sig_w_arr(4)

      w_av_p = w_av

      IF ( humidity )  THEN
         sig_q_arr(1) = sig_q_arr(2)
         sig_q_arr(2) = sig_q_arr(3)
         sig_q_arr(3) = sig_q_arr(4)

         q_av_p = q_av
      ENDIF

      IF ( .NOT. neutral )  THEN
         sig_pt_arr(1) = sig_pt_arr(2)
         sig_pt_arr(2) = sig_pt_arr(3)
         sig_pt_arr(3) = sig_pt_arr(4)

         pt_av_p = pt_av

      ENDIF


         count_time_u = count_time_u + 1
         count_time_v = count_time_v + 1
         count_time_w = count_time_w + 1
         IF ( humidity ) count_time_q = count_time_q + 1
         IF ( .NOT. neutral )  count_time_pt = count_time_pt + 1



   END SUBROUTINE vdi_domain_averages


!------------------------------------------------------------------------------!
! Description:
! ------------
!>
!------------------------------------------------------------------------------!
   SUBROUTINE vdi_conservation_of_mass

      USE arrays_3d

      USE control_parameters

      USE indices

      USE kinds

      USE pegrid

      USE grid_variables


      IMPLICIT NONE

      INTEGER(iwp) ::  i
      INTEGER(iwp) ::  j
      INTEGER(iwp) ::  k

      REAL(wp) ::  d_area_t                        !< inverse of the total area of the horizontal model domain
      REAL(wp) ::  w_correct
      REAL(wp), DIMENSION(1:3) ::  volume_flow_l


      volume_flow = 0.0_wp
      volume_flow_l = 0.0_wp



!
!--   Left/right
      volume_flow(1)   = 0.0_wp
      volume_flow_l(1) = 0.0_wp

      IF ( nxl == 0 )  THEN
         i = 0
         DO  j = nys, nyn
!
!--         Sum up the volume flow through the left boundary
            DO  k = nzb+1, nzt
               volume_flow_l(1) = volume_flow_l(1) + u(k,j,i) * dzw(k) * dy      &
                                     * MERGE( 1.0_wp, 0.0_wp,                  &
                                              BTEST( wall_flags_0(k,j,i), 1 )  &
                                            )
            ENDDO
         ENDDO
      ENDIF


      IF ( nxr == nx)  THEN
         i = nx+1
         DO  j = nys, nyn
!
!--      Sum up the volume flow through the right boundary
            DO  k = nzb+1, nzt
               volume_flow_l(1) = volume_flow_l(1) - u(k,j,i) * dzw(k) * dy    &
                                     * MERGE( 1.0_wp, 0.0_wp,                  &
                                              BTEST( wall_flags_0(k,j,i), 1 )  &
                                            )
            ENDDO
         ENDDO
      ENDIF

#if defined( __parallel )
      IF ( collective_wait )  CALL MPI_BARRIER( comm1dy, ierr )
      CALL MPI_ALLREDUCE( volume_flow_l(1), volume_flow(1), 1, MPI_REAL,       &
                           MPI_SUM, comm1dy, ierr )
#else
      volume_flow = volume_flow_l
#endif


!
!--   South/north
      volume_flow(2)   = 0.0_wp
      volume_flow_l(2) = 0.0_wp

      IF ( nys == 0 )  THEN
         j = 0
         DO  i = nxl, nxr
!
!--         Sum up the volume flow through the south boundary
            DO  k = nzb+1, nzt
               volume_flow_l(2) = volume_flow_l(2) + v(k,j,i) * dzw(k) * dx    &
                                     * MERGE( 1.0_wp, 0.0_wp,                  &
                                              BTEST( wall_flags_0(k,j,i), 2 )  &
                                            )
            ENDDO
         ENDDO
      ENDIF


      IF ( nyn == ny)  THEN
         j = ny+1
         DO  i = nxl, nxr
!
!--         Sum up the volume flow through the north boundary
            DO  k = nzb+1, nzt
               volume_flow_l(2) = volume_flow_l(2) - v(k,j,i) * dzw(k) * dx    &
                                     * MERGE( 1.0_wp, 0.0_wp,                  &
                                              BTEST( wall_flags_0(k,j,i), 2 )  &
                                            )
            ENDDO
         ENDDO
      ENDIF

#if defined( __parallel )
      IF ( collective_wait )  CALL MPI_BARRIER( comm1dy, ierr )
      CALL MPI_ALLREDUCE( volume_flow_l(1), volume_flow(1), 1, MPI_REAL,       &
                           MPI_SUM, comm1dy, ierr )
#else
      volume_flow = volume_flow_l
#endif



!
!--   Top boundary
      volume_flow(3) = 0.0_wp
      volume_flow_l(3) = 0.0_wp
      k = nzt
      DO  i = nxl, nxr
         DO  j = nys, nyn
            volume_flow_l(3) = volume_flow_l(3) - w(k,j,i) * dx * dy
         ENDDO
      ENDDO

#if defined( __parallel )
      IF ( collective_wait )  CALL MPI_BARRIER( comm2d, ierr )
      CALL MPI_ALLREDUCE( volume_flow_l, volume_flow, 3, MPI_REAL, MPI_SUM,    &
                          comm2d, ierr )
#else
      volume_flow = volume_flow_l
#endif



      d_area_t = 1.0_wp / ( ( nx + 1 ) * dx * ( ny + 1 ) * dy )
      w_correct = SUM ( volume_flow) * d_area_t


      IF ( ABS( w_correct ) > 0.001 )  THEN
         message_string = 'The mass is not conserved. '
         PRINT*, w_correct
         CALL message( 'vdi_conservation_of_mass', 'PAXXXX', 0, 2, 0, 6, 0 )
      ENDIF



   END SUBROUTINE vdi_conservation_of_mass



!------------------------------------------------------------------------------!
! Description:
! ------------
!> The results will be checked for exceedance the specified limits.
!> The controls are performed at every time step and at every grid point.
!> No wind component is allowed to have a magnitude greater than ten times
!> the maximum wind velocity at the approach flow profile (vdi 3783 part 9).
!------------------------------------------------------------------------------!
   SUBROUTINE vdi_plausible_values

      USE arrays_3d

      USE control_parameters

      USE indices

      USE kinds

      USE pegrid


      IMPLICIT NONE

      INTEGER(iwp) ::  i
      INTEGER(iwp) ::  j
      INTEGER(iwp) ::  k


      REAL(wp), DIMENSION(:), ALLOCATABLE ::  uv
      REAL(wp), DIMENSION(:), ALLOCATABLE ::  uv_l


      IF ( .NOT. ALLOCATED( uv ) )  THEN
        ALLOCATE( uv(nzb:nzt+1) )
        ALLOCATE( uv_l(nzb:nzt+1) )

        uv   = 0.0_wp
        uv_l = 0.0_wp
      ENDIF


!
!--   Determination of the approach flow profile
      IF ( bc_lr_cyc  .AND.  bc_ns_cyc )  THEN
         IF ( nxl == 0 .AND. nys == 0 )  THEN
         DO  k = nzb, nzt+1
            uv_l(k) = SQRT( ( 0.5_wp * ( u(k,0,-1) + u(k,0,0) ) )**2 +         &
                           ( 0.5_wp * ( v(k,-1,0) + v(k,0,0) ) )**2 )
         ENDDO
         ENDIF
      ENDIF


      IF ( bc_dirichlet_l  .OR.  bc_radiation_l )  THEN
         IF ( nxl == 0 .AND. nys == 0 )  THEN
            uv_l(k) = SQRT( ( 0.5_wp * ( u(k,0,-1) + u(k,0,0) ) )**2 +         &
                         ( 0.5_wp * ( v(k,-1,0) + v(k,0,0) ) )**2 )
         ENDIF
      ENDIF

      IF ( bc_dirichlet_r  .OR.  bc_radiation_r )  THEN
         IF ( nxr == nx .AND. nys == 0 )  THEN
            uv_l(k) = SQRT( ( 0.5_wp * ( u(k,0,nxr) + u(k,0,nxr+1) ) )**2 +    &
                         ( 0.5_wp * ( v(k,-1,nxr) + v(k,0,nxr) ) )**2 )
         ENDIF
      ENDIF



      IF ( bc_dirichlet_n  .OR.  bc_radiation_n )  THEN
         IF ( nxl == 0 .AND. nyn == ny )  THEN
            uv_l(k) = SQRT( ( 0.5_wp * ( u(k,nyn,-1) + u(k,nyn,0) ) )**2 +     &
                         ( 0.5_wp * ( v(k,nyn+1,0) + v(k,nyn,0) ) )**2 )
         ENDIF
      ENDIF


      IF ( bc_dirichlet_s  .OR.  bc_radiation_s )  THEN
         IF ( nxl == 0 .AND. nys == 0 )  THEN
            uv_l(k) = SQRT( ( 0.5_wp * ( u(k,0,-1) + u(k,0,0) ) )**2 +         &
                           ( 0.5_wp * ( v(k,-1,0) + v(k,0,0) ) )**2 )
         ENDIF
      ENDIF




#if defined( __parallel )
      CALL MPI_ALLREDUCE( uv_l, uv, nzt+1-nzb+1,                               &
                                 MPI_REAL, MPI_MAX, comm2d, ierr )
#else
      uv = uv_l
#endif

!
!--   Test for exceedance the specified limits
      DO  i = nxl, nxr
         DO  j = nys, nyn
            DO  k = nzb+1, nzt
               message_string = 'A wind component have a magnitude greater ' //&
                                'than ten times the maximum wind velocity ' // &
                                'at the approach flow profile.'


               IF ( ABS(u(k,j,i)) > 10.0_wp * MAXVAL(uv) )  THEN
!                   PRINT*, u(k,j,i), 10.0_wp * MAXVAL(uv), 'u', k
                  CALL message( 'vdi_plausible_values', 'PAXXXX', 0, 2, myid, 6, 0 )
               ENDIF

               IF ( ABS(v(k,j,i)) > 10.0_wp * MAXVAL(uv) )  THEN
!                                  PRINT*, v(k,j,i), 10.0_wp * MAXVAL(uv), 'v'
                  CALL message( 'vdi_plausible_values', 'PAXXXX', 0, 2, myid, 6, 0 )
               ENDIF

               IF ( ABS(w(k,j,i)) > 10.0_wp * MAXVAL(uv) )  THEN
!                                  PRINT*, w(k,j,i), 10.0_wp * MAXVAL(uv), 'w'
                  CALL message( 'vdi_plausible_values', 'PAXXXX', 0, 2, myid, 6, 0 )
               ENDIF

            ENDDO
         ENDDO
      ENDDO

!
!--  Test if the potential temperature lies between 220 K and 330 K
     IF ( MAXVAL(pt) > 330.0_wp .or. MAXVAL(pt) < 220.0_wp ) THEN
        message_string = 'The potential temperature does not lie ' //          &
                         'between 220 K and 330 K.'
        CALL message( 'vdi_plausible_values', 'PAXXXX', 0, 2, myid, 6, 0 )
     ENDIF


!
!--  The supersaturation can not more than 10 %.  Therefore, no test is required.


   END SUBROUTINE vdi_plausible_values











 END MODULE vdi_internal_controls
